from __future__ import annotations

import os
import json
from dataclasses import dataclass, field, asdict
from typing import (
    TYPE_CHECKING,
    Sequence,
    Union,
    Literal,
    Optional,
    List,
    overload,
    Any,
    Type,
)

from norminette.colors import error_color
from norminette.norm_error import errors as errors_dict

if TYPE_CHECKING:
    from norminette.lexer import Token
    from norminette.file import File

ErrorLevel = Literal["Error", "Notice"]


@dataclass
class Highlight:
    lineno: int
    column: int
    length: Optional[int] = field(default=None)
    hint: Optional[str] = field(default=None)

    @classmethod
    def from_token(
        cls,
        token: Token,
        *,
        hint: Optional[str] = None,
    ) -> Highlight:
        return cls(token.lineno, token.column, token.unsafe_length, hint)

    def __lt__(self, other: Any) -> bool:
        assert isinstance(other, Highlight)
        if self.lineno == other.lineno:
            if self.column == other.column:
                return len(self.hint or '') > len(other.hint or '')
            return self.column > other.column
        return self.lineno > other.lineno


@dataclass
class Error:
    name: str
    text: str
    level: ErrorLevel = field(default="Error")
    highlights: List[Highlight] = field(default_factory=list)

    @classmethod
    def from_name(cls: Type[Error], /, name: str, **kwargs) -> Error:
        return cls(name, errors_dict[name], **kwargs)

    def __lt__(self, other: Any) -> bool:
        assert isinstance(other, Error)
        if not self.highlights:
            return bool(other.highlights) or self.name > other.name
        if not other.highlights:
            return bool(self.highlights) or other.name > self.name
        ah, bh = min(self.highlights), min(other.highlights)
        if ah.column == bh.column and ah.lineno == bh.lineno:
            return self.name < other.name
        return (ah.lineno, ah.column) < (bh.lineno, bh.column)

    @overload
    def add_highlight(
        self,
        lineno: int,
        column: int,
        length: Optional[int] = None,
        hint: Optional[str] = None,
    ) -> None: ...
    @overload
    def add_highlight(self, highlight: Highlight, /) -> None: ...

    def add_highlight(self, *args, **kwargs) -> None:
        if len(args) == 1:
            highlight, = args
        else:
            highlight = Highlight(*args, **kwargs)
        self.highlights.append(highlight)


class Errors:
    __slots__ = "_inner"

    def __init__(self) -> None:
        self._inner: List[Error] = []

    def __repr__(self) -> str:
        return repr(self._inner)

    def __len__(self) -> int:
        return len(self._inner)

    def __iter__(self):
        self._inner.sort()
        return iter(self._inner)

    @overload
    def add(self, error: Error) -> None:
        """Add an `Error` instance to the errors.
        """
        ...

    @overload
    def add(self, name: str, *, level: ErrorLevel = "Error", highlights: List[Highlight] = ...) -> None:
        """Builds an `Error` instance from a name in `errors_dict` and adds it to the errors.

        ```python
        >>> errors.add("TOO_MANY_LINES")
        >>> errors.add("INVALID_HEADER")
        >>> errors.add("GLOBAL_VAR_DETECTED", level="Notice")
        ```
        """
        ...

    @overload
    def add(
        self,
        /,
        name: str,
        text: str,
        *,
        level: ErrorLevel = "Error",
        highlights: List[Highlight] = ...,
    ) -> None:
        """Builds an `Error` instance and adds it to the errors.

        ```python
        >>> errors.add("BAD_IDENTATION", "You forgot an column here")
        >>> errors.add("CUSTOM_ERROR", f"name {not_defined!r} is not defined. Did you mean: {levenshtein_distance}?")
        >>> errors.add("NOOP", "Empty if statement", level="Notice")
        ```
        """
        ...

    def add(self, *args, **kwargs) -> None:
        kwargs.setdefault("level", "Error")
        error = None
        if len(args) == 1:
            error = args[0]
            if isinstance(error, str):
                error = Error.from_name(error, **kwargs)
        if len(args) == 2:
            error = Error(*args, **kwargs)
        assert isinstance(error, Error), "bad function call"
        return self._inner.append(error)

    @property
    def status(self) -> Literal["OK", "Error"]:
        return "OK" if all(it.level == "Notice" for it in self._inner) else "Error"

    def append(self, *args, **kwargs):
        """Deprecated alias for `.add(...)`, kept for backward compatibility.

        Use `.add(...)` instead.
        """
        return self.add(*args, **kwargs)


class _formatter:
    name: str

    def __init__(self, files: Union[File, Sequence[File]], **options) -> None:
        if not isinstance(files, Sequence):
            files = [files]
        self.files = files
        self.options = options

    def __init_subclass__(cls) -> None:
        cls.name = cls.__name__.rstrip("ErrorsFormatter").lower()


class HumanizedErrorsFormatter(_formatter):
    @property
    def use_colors(self) -> bool:
        return self.options.get("use_colors", True)

    def _colorize_error_text(self, error: Error) -> str:
        color = error_color(error.name)
        if not self.use_colors or not color:
            return error.text
        return f"\x1b[{color}m{error.text}\x1b[0m"

    def __str__(self) -> str:
        output = ''
        for file in self.files:
            output += f"{file.basename}: {file.errors.status}!"
            for error in file.errors:
                highlight = error.highlights[0]
                error_text = self._colorize_error_text(error)
                output += f"\n{error.level}: {error.name:<20} "
                output += f"(line: {highlight.lineno:>3}, col: {highlight.column:>3}):\t{error_text}"
            output += '\n'
        return output


class JSONErrorsFormatter(_formatter):
    def __str__(self):
        files = []
        for file in self.files:
            files.append({
                "path": os.path.abspath(file.path),
                "status": file.errors.status,
                "errors": tuple(map(asdict, file.errors)),
            })
        output = {
            "files": files,
        }
        return json.dumps(output, separators=(',', ':')) + '\n'


formatters = (
    JSONErrorsFormatter,
    HumanizedErrorsFormatter,
)



import glob
import sys
import pathlib
import platform
from importlib.metadata import version

import argparse
from norminette.errors import formatters
from norminette.file import File
from norminette.lexer import Lexer
from norminette.exceptions import CParsingError
from norminette.registry import Registry
from norminette.context import Context
from norminette.tools.colors import colors

import subprocess

version_text = f"norminette {version('norminette')}"
version_text += f", Python {platform.python_version()}"
version_text += f", {platform.platform()}"


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "file",
        help="File(s) or folder(s) you wanna run the parser on. If no file provided, runs on current folder.",
        nargs="*",
    )
    parser.add_argument(
        "-d",
        "--debug",
        action="count",
        help="Debug output (-dd outputs the whole tokenization and such, used for developping)",
        default=0,
    )
    parser.add_argument(
        "-o",
        "--only-filename",
        action="store_true",
        help="By default norminette displays the full path to the file, this allows to show only filename",
        default=False,
    )
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=version_text,
    )
    parser.add_argument(
        "--cfile",
        action="store",
        help="Store C file content directly instead of filename",
    )
    parser.add_argument(
        "--hfile",
        action="store",
        help="Store header file content directly instead of filename",
    )
    parser.add_argument(
        "--filename",
        action="store",
        help="Stores filename if --cfile or --hfile is passed",
    )
    parser.add_argument(
        "--use-gitignore",
        action="store_true",
        help="Parse only source files not match to .gitignore",
    )
    parser.add_argument(
        "-f",
        "--format",
        choices=list(formatter.name for formatter in formatters),
        help="formatting style for errors",
        default="humanized",
    )
    parser.add_argument(
        "--no-colors", action="store_true", help="Disable colors in output"
    )
    parser.add_argument("-R", nargs=1, help="compatibility for norminette 2")
    args = parser.parse_args()
    registry = Registry()

    format = next(filter(lambda it: it.name == args.format, formatters))
    files = []
    debug = args.debug
    if args.cfile or args.hfile:
        file_name = args.filename or ("file.c" if args.cfile else "file.h")
        file_data = args.cfile if args.cfile else args.hfile
        file = File(file_name, file_data)
        files.append(file)
    else:
        stack = []
        stack += args.file if args.file else glob.glob("**/*.[ch]", recursive=True)
        for item in stack:
            path = pathlib.Path(item)
            if not path.exists():
                print(f"Error: '{path!s}' no such file or directory")
                sys.exit(1)
            if path.is_file():
                if path.suffix not in (".c", ".h"):
                    print(f"Error: {path.name!r} is not valid C or C header file")
                else:
                    file = File(item)
                    files.append(file)
            if path.is_dir():
                stack += glob.glob(str(path) + "/**/*.[ch]", recursive=True)
        del stack

    if args.use_gitignore:
        tmp_targets = []
        for target in files:
            command = ["git", "check-ignore", "-q", target.path]
            exit_code = subprocess.run(
                command, stdout=subprocess.PIPE, stderr=subprocess.PIPE
            ).returncode
            """
            see: $ man git-check-ignore
            EXIT STATUS
                  0: One or more of the provided paths is ignored.
                  1: None of the provided paths are ignored.
                128: A fatal error was encountered.
            """
            if exit_code == 0:
                pass
            elif exit_code == 1:
                tmp_targets.append(target)
            elif exit_code == 128:
                print(
                    f"Error: something wrong with --use-gitignore option {target.path!r}"
                )
                sys.exit(0)
        files = tmp_targets
    for file in files:
        try:
            lexer = Lexer(file)
            tokens = list(lexer)
            context = Context(file, tokens, debug, args.R)
            registry.run(context)
        except CParsingError as e:
            print(file.path + f": Error!\n\t{colors(e.msg, 'red')}")
            sys.exit(1)
        except KeyboardInterrupt:
            sys.exit(1)
    errors = format(files, use_colors=not args.no_colors)
    print(errors, end="")
    sys.exit(1 if len(file.errors) else 0)


if __name__ == "__main__":
    main()



import importlib
import os
from operator import attrgetter

from norminette.rules.rule import Rule, Primary, Check


class Rules:
    __slots__ = (
        "all",
        "primaries",
        "checks",
    )

    __instance = None

    def __new__(cls):
        if not cls.__instance:
            cls.__instance = super().__new__(cls)
        return cls.__instance

    def __init__(self) -> None:
        path = os.path.dirname(os.path.realpath(__file__))
        for f in os.listdir(path):
            name, _ = os.path.splitext(f)
            importlib.import_module("norminette.rules." + name)

        self.all = Rule.__subclasses__()
        self.checks = Check.__subclasses__()
        self.primaries = sorted(Primary.__subclasses__(), reverse=True, key=attrgetter("priority"))



import collections
from operator import attrgetter

from norminette.rules import Rules, Primary
from norminette.exceptions import CParsingError

rules = Rules()


class Registry:
    global has_err

    def __init__(self):
        self.dependencies = collections.defaultdict(list)
        for rule in rules.checks:
            rule.register(self)
        for name, dependencies in self.dependencies.items():
            self.dependencies[name] = sorted(dependencies, reverse=True, key=attrgetter("__name__"))

    def run_rules(self, context, rule):
        rule = rule(context)
        result = rule.run(context)
        ret, read = result if isinstance(rule, Primary) else (False, 0)
        if ret:
            context.scope.instructions += 1
            if isinstance(rule, Primary):
                context.tkn_scope = read
                context.history.append(rule)
            for rule in self.dependencies[rule.name]:
                self.run_rules(context, rule)
            for rule in self.dependencies["_rule"]:
                self.run_rules(context, rule)
            context.tkn_scope = 0
        return ret, read

    def run(self, context):
        """
        Main function for each file.
        Primary rules are determined by the prefix "Is" and
        are run by order of priority as defined in each class
        Each secondary rule is then run in arbitrary order based on their
        dependencies
        """
        unrecognized_tkns = []
        context.state = "starting"
        for rule in self.dependencies["_start"]:
            self.run_rules(context, rule)
        context.state = "running"
        while context.tokens != []:
            context.tkn_scope = len(context.tokens)
            for rule in rules.primaries:
                if rule.scope and context.scope not in rule.scope:
                    continue
                ret, jump = self.run_rules(context, rule)
                if ret is True:
                    if unrecognized_tkns != []:
                        if context.debug == 0:
                            raise CParsingError(
                                f"Error: Unrecognized line {unrecognized_tkns[0].pos} while parsing line {unrecognized_tkns}"  # noqa: E501
                            )
                        print("uncaught -> ", context.file.name)
                        print("uncaught -> ", unrecognized_tkns)
                        unrecognized_tkns = []
                    context.dprint(rule.name, jump)
                    context.update()
                    context.pop_tokens(jump)
                    break
            # #############################################################
            else:  # Remove these one ALL  primary rules are done
                # print("#, ", context.tokens[0])
                unrecognized_tkns.append(context.tokens[0])
                context.pop_tokens(1)  # ##################################
            # #############################################################
        context.state = "ending"
        for rule in self.dependencies["_end"]:
            self.run_rules(context, rule)
        if unrecognized_tkns != []:
            print(context.debug)
            if context.debug > 0:
                print("uncaught ->", unrecognized_tkns)



from norminette.scope import Function
from norminette.context import GlobalScope
from norminette.rules import Rule, Primary


whitespaces = ["SPACE", "TAB"]

SEPARATORS = ["COMMA", "AND", "OR", "SEMI_COLON"]
assigns = [
    "RIGHT_ASSIGN",
    "LEFT_ASSIGN",
    "ADD_ASSIGN",
    "SUB_ASSIGN",
    "MUL_ASSIGN",
    "DIV_ASSIGN",
    "MOD_ASSIGN",
    "AND_ASSIGN",
    "XOR_ASSIGN",
    "OR_ASSIGN",
    "ASSIGN",
]
misc_identifier = [
    "CONST",
    "REGISTER",
    "STATIC",
    "VOLATILE",
    "EXTERN",
    "INLINE",
    "RESTRICT",
    "SIGNED",
    "UNSIGNED",
    "TYPEDEF",
    "STRUCT",
    "ENUM",
    "UNION",
]
type_identifier = [
    "CHAR",
    "DOUBLE",
    "ENUM",
    "FLOAT",
    "INT",
    "UNION",
    "VOID",
    "LONG",
    "SHORT",
]


class IsFuncDeclaration(Rule, Primary, priority=81):
    scope = (
        GlobalScope,
    )

    def check_args(self, context, pos):
        i = context.skip_ws(pos, nl=True)
        while context.check_token(i, whitespaces + ["RPARENTHESIS"]):
            i += 1
        if context.check_token(i, "LPARENTHESIS") is False:
            return False, pos
        p = 1
        i += 1
        while p:
            if context.check_token(i, "LPARENTHESIS"):
                p += 1
            elif context.check_token(i, "RPARENTHESIS"):
                p -= 1
            elif context.peek_token(i) is None:
                return False, 0
            i += 1
        return True, i

    def check_func_identifier(self, context, pos):
        i = context.skip_ws(pos, nl=True)
        pp = 0  # Pointer operator's position
        lp = 0  # Left parenthesis counter (nesting level)

        if context.check_token(i, "IDENTIFIER"):
            i += 1
            return True, i, False

        d = ["LPARENTHESIS", "MULT"] + whitespaces
        while context.check_token(i, d):
            if context.check_token(i, "MULT") and not pp:
                pp = i
            elif pp and context.check_token(i, "LPARENTHESIS"):
                lp += 1
            i += 1
        i = context.skip_misc_specifier(i, nl=True)
        if context.check_token(i, "IDENTIFIER") is False:
            return False, pos, False

        i += 1
        while context.check_token(i, ["RPARENTHESIS"] + whitespaces):
            if context.check_token(i, "RPARENTHESIS"):
                lp -= 1
            i += 1
        if pp and lp < 0 and context.check_token(i, "LPARENTHESIS"):
            return False, pos, False

        return True, i, (True if pp else False)

    def check_func_format(self, context):
        i = context.skip_ws(0, nl=False)
        type_id = []
        misc_id = []
        identifier = None
        args = False
        if context.check_token(i, "NEWLINE") is True:
            return False, 0
        while context.peek_token(i):
            while (
                context.check_token(i, "IDENTIFIER") is True
                and context.peek_token(i).value == "__attribute__"
            ):
                i += 1
                i = context.skip_ws(i)
                i = context.skip_nest(i)
                i = context.skip_ws(i)
            if (
                context.check_token(i, "NEWLINE") is True
                and identifier is False
                and misc_id == []
                and type_id == []
            ):
                return False, 0
            if context.check_token(i, misc_identifier) is True:
                misc_id.append(context.peek_token(i))
            elif context.check_token(i, type_identifier) is True:
                type_id.append(context.peek_token(i))
            if (
                context.check_token(
                    i, assigns + ["TYPEDEF", "COMMA", "LBRACE", "HASH"]
                )
                is True
            ):
                return False, 0
            if context.check_token(i, "SEMI_COLON") is True:
                return False, 0
            elif context.check_token(i, "IDENTIFIER") is True:
                if identifier is not None:
                    type_id.append(identifier[0])
                identifier = (context.peek_token(i), i)
            if context.check_token(i, "NEWLINE") is True:
                if args is False:
                    i += 1
                    continue
                else:
                    break
            if context.check_token(i, "LPARENTHESIS") is True:
                par = context.parenthesis_contain(i)
                if par[0] == "function":
                    if identifier is not None:
                        type_id.append(identifier[0])
                    while context.check_token(i, ["LPARENTHESIS", "MULT", "BWISE_AND"]):
                        i += 1
                    identifier = (context.peek_token(i), i)
                    i = context.skip_nest(i)
                elif identifier is not None:
                    args = True
                    arg_start = i
                    i = context.skip_nest(i)
                    arg_end = i
                    while (context.check_token(i, "RPARENTHESIS")) is True:
                        i += 1
                    i = context.skip_ws(i)
                    if context.check_token(i, "LPARENTHESIS") is True:
                        arg_start = i
                        i = context.skip_nest(i)
                        arg_end = i
                    break
                else:
                    i += 1
            else:
                i += 1
        if len(type_id) > 0 and args is True and identifier is not None:
            i = identifier[1]
            i = context.skip_ws(i, nl=True)
            while context.check_token(i, ["LPARENTHESIS", "MULT", "BWISE_AND"]) is True:
                i += 1
            sc = context.scope
            while type(sc) != GlobalScope:
                sc = sc.outer()
            sc.fnames.append(context.peek_token(i).value)
            context.fname_pos = i
            context.arg_pos = [arg_start, arg_end]
            i = arg_end
            i = context.skip_ws(i, nl=True)
            while context.check_token(i, ["RPARENTHESIS"]) is True:
                i += 1
            i = context.skip_ws(i, nl=True)
            if context.check_token(i, "LPARENTHESIS") is True:
                i = context.skip_nest(i)
                i += 1
            i = context.skip_ws(i, nl=True)
            while context.check_token(i, "LBRACKET"):
                i = context.skip_nest(i)
                i += 1
                i = context.skip_ws(i, nl=True)
            while (
                context.check_token(i, "IDENTIFIER") is True
                and context.peek_token(i).value == "__attribute__"
            ):
                i += 1
                i = context.skip_ws(i)
                i = context.skip_nest(i) + 1
                i = context.skip_ws(i)
            return True, i
        return False, 0

    def run(self, context):
        """
        Catches function declaration
        Allows newline inside it
        Creates context variable for function name, arg_start, arg_end
        """
        if type(context.scope) is not GlobalScope:
            return False, 0
        ret, read = self.check_func_format(context)
        if ret is False:
            return False, 0
        while context.check_token(read, ["COMMENT", "MULT_COMMENT"]) is True:
            read += 1
        read = context.skip_ws(read, nl=False)
        if context.check_token(read, ["NEWLINE", "LBRACE", "HASH"]):
            if context.check_token(read, ["LBRACE", "HASH"]) is True:
                read -= 1
            context.scope.functions += 1
            read += 1
            temp = context.skip_ws(read, nl=True)
            if context.check_token(temp, "LBRACE"):
                context.sub = context.scope.inner(Function)
            read = context.eol(read)
            return True, read

        elif context.check_token(read, SEPARATORS):
            read += 1
            read = context.eol(read)
            return False, 0

        return False, 0



from norminette.rules import Rule, Check


class CheckFunctionsCount(Rule, Check):
    depends_on = (
        "IsFuncDeclaration",
    )

    def run(self, context):
        """
        Each file cannot contain more than 5 function
        """
        if context.scope is not None and context.scope.name == "GlobalScope":
            if context.scope.functions > 5:
                context.new_error("TOO_MANY_FUNCS", context.peek_token(0))
        return False, 0



import sys
import contextlib

from norminette.rules import Rule, Primary
from norminette.lexer.dictionary import keywords
from norminette.exceptions import CParsingError
from norminette.context import Macro

UNARY_OPERATORS = (
    "PLUS",
    "MINUS",
    "NOT",  # !
    "BWISE_NOT",  # ~
)

BINARY_OPERATORS = (
    # Arithmetic operators
    "PLUS",
    "MINUS",
    "MULT",
    "DIV",
    "MODULO",
    # Relational operators
    "EQUALS",  # ==
    "NOT_EQUAL",  # !=
    "MORE_THAN",  # > (why not GREATER_THAN?)
    "LESS_THAN",  # <
    "GREATER_OR_EQUAL",  # >=
    "LESS_OR_EQUAL",  # <=
    # Logical operators
    "AND",
    "OR",
    # Bitwise operators
    "BWISE_AND",  # &
    "BWISE_OR",  # |
    "BWISE_XOR",  # ^
    "BWISE_XOR",  # ^
    "LEFT_SHIFT",  # << (why not BWISE_LEFT?)
    "RIGHT_SHIFT",  # >>
)

ALLOWED_IN_PATH = (
    "IDENTIFIER",
    "DIV",
    "MINUS",
    "DOT",
    "SPACE",
    "TAB",
    # TODO Remove all keyword tokens and add to just use 'IDENTIFIER' instead
    *keywords.values(),  # https://github.com/42School/norminette/issues/470
)


@contextlib.contextmanager
def recursion_limit(limit):
    old_limit = sys.getrecursionlimit()
    sys.setrecursionlimit(limit)
    yield
    sys.setrecursionlimit(old_limit)


class IsPreprocessorStatement(Rule, Primary, priority=100):
    def run(self, context):
        """
        Catches any kind of preprocessor statements
        Handles indentation related informations
        """
        i = context.skip_ws(0)
        if not context.check_token(i, "HASH"):
            return False, 0
        self.hash = context.peek_token(i)
        i += 1
        i = context.skip_ws(i)
        if context.check_token(i, "NEWLINE"):  # Null directive
            return True, i + 1                 # TODO: Fix null directives (comments)
        # Why `if` and `else` need to be a special case?
        if not context.check_token(i, ("IDENTIFIER", "IF", "ELSE")):
            raise CParsingError(f"Invalid preprocessor statement {context.peek_token(i)}")
        token = context.peek_token(i)
        direc = (token.value if token.type == "IDENTIFIER" else token.type).lower()
        i += 1
        i = context.skip_ws(i)
        if checker := getattr(self, f"check_{direc}", None):
            return checker(context, i)
        raise CParsingError(f"Invalid preprocessing directive #{direc}")

    def check_define(self, context, index):
        if not context.check_token(index, "IDENTIFIER"):
            raise CParsingError("No identifier after #define")
        token = context.peek_token(index)
        index += 1
        if is_function := context.check_token(index, "LPARENTHESIS"):
            index += 1
            index = context.skip_ws(index)
            while context.check_token(index, "IDENTIFIER"):
                index += 1
                index = context.skip_ws(index)
                if context.check_token(index, "COMMA"):
                    index += 1
                    index = context.skip_ws(index)
            # Add better errors like check EOF and invalid identifier?
            if not context.check_token(index, "RPARENTHESIS"):
                raise CParsingError("Invalid macro function definition")
            index += 1
        macro = Macro.from_token(token, is_func=is_function)
        context.preproc.macros.append(macro)
        index = context.skip_ws(index)
        return self._just_token_string("define", context, index)

    def check_import(self, context, index):
        is_valid_argument, index = self._check_path(context, index)
        if not is_valid_argument:
            raise CParsingError("Invalid file argument for #import directive")
        index = context.skip_ws(index)
        return self._just_token_string("import", context, index)

    def check_pragma(self, context, index):
        return self._just_token_string("pragma", context, index)

    def check_error(self, context, index):
        return self._just_token_string("error", context, index)

    def check_warning(self, context, index):
        return self._just_token_string("warning", context, index)

    def check_if(self, context, index):
        if not self.corresponding_endif(context, index):
            context.new_error("PREPROC_BAD_IF", self.hash)
        context.preproc.indent += 1
        context.preproc.total_ifs += 1
        return self._just_constant_expression("if", context, index)

    def check_elif(self, context, index):
        if not self.corresponding_endif(context, index):
            context.new_error("PREPROC_BAD_ELIF", self.hash)
        context.preproc.total_elifs += 1
        return self._just_constant_expression("elif", context, index)

    def check_ifdef(self, context, index):
        if not self.corresponding_endif(context, index):
            context.new_error("PREPROC_BAD_IFDEF", self.hash)
        context.preproc.indent += 1
        context.preproc.total_ifdefs += 1
        return self._just_identifier("ifdef", context, index)

    def check_ifndef(self, context, index):
        if not self.corresponding_endif(context, index):
            context.new_error("PREPROC_BAD_IFNDEF", self.hash)
        context.preproc.indent += 1
        context.preproc.total_ifndefs += 1
        return self._just_identifier("infdef", context, index)

    def check_undef(self, context, index):
        return self._just_identifier("undef", context, index)

    def check_else(self, context, index):
        if context.preproc.indent == 0:
            context.new_error("PREPROC_BAD_ELSE", self.hash)
        context.preproc.total_elses += 1
        return self._just_eol("else", context, index)

    def check_endif(self, context, index):
        if context.preproc.indent == 0:
            context.new_error("PREPROC_BAD_ENDIF", self.hash)
        context.preproc.indent -= 1
        return self._just_eol("endif", context, index)

    def check_include(self, context, index):
        is_valid_argument, index = self._check_path(context, index)
        if not is_valid_argument:
            raise CParsingError("Invalid file argument for #include directive")
        return self._just_eol("include", context, index)

    def corresponding_endif(self, context, index):
        """Checks if the corresponding `#endif` is present.
        """
        depth = 0
        while index < len(context.tokens):
            if not context.check_token(index, "HASH"):
                index += 1
                continue

            index += 1
            index = context.skip_ws(index)
            if not context.check_token(index, ("IDENTIFIER", "IF", "ELSE")):
                continue

            token = context.peek_token(index)
            direc = (token.value if token.type == "IDENTIFIER" else token.type).lower()
            if direc == "endif":
                if depth == 0:
                    return True
                depth -= 1
            elif direc in ("if", "ifdef", "ifndef"):
                depth += 1
            index += 1
        return False

    def _check_path(self, context, index):
        """Checks the argument of an include/import statement.

        Examples of valid headers:
        - `"libft.h"`
        - `<    bla.h   >`
        - `<42.h   >`
        - `<   four.two>`
        """
        # TODO: It not works with `#include <if.h>` because of the `if` keyword
        if context.check_token(index, "STRING"):
            index += 1
            return True, index
        if not context.check_token(index, "LESS_THAN"):
            return False, index
        index = context.skip_ws(index + 1)
        while context.check_token(index, ALLOWED_IN_PATH):
            index += 1
        if not context.check_token(index, "MORE_THAN"):
            return False, index
        index += 1
        return True, index

    def _just_token_string(self, directive, context, index):
        index = context.skip_ws(index, comment=True)
        if context.check_token(index, "NEWLINE"):
            index += 1
            return True, index
        lines = 1
        newline = False
        while context.peek_token(index) is not None and lines > 0:
            if context.check_token(index, "NEWLINE"):
                lines -= 1
                newline = False
            elif context.check_token(index, "BACKSLASH") and not newline:
                lines += 1
                newline = True
            index += 1
        if lines > 0 and context.peek_token(index) is not None:
            raise CParsingError(f"Unexpected end of file after #{directive} directive")
        return True, index

    def _just_constant_expression(self, directive, context, index):
        parser = ConstantExpressionParser(directive, context, index)
        ok, index = parser.parse()
        if not ok:
            return ok, index
        return self._just_eol(directive, context, index)

    def _just_identifier(self, directive, context, index):
        if not context.check_token(index, "IDENTIFIER"):
            raise CParsingError(f"Invalid argument for #{directive} statement")
        index += 1
        return self._just_eol(directive, context, index)

    def _just_eol(self, directive, context, index):
        index = context.skip_ws(index, comment=True)
        if context.peek_token(index) is None:
            return True, index
        #     raise CParsingError(f"Unexpected end of file after #{directive} directive")
        if not context.check_token(index, "NEWLINE"):
            raise CParsingError(f"Extra tokens at end of #{directive} directive")
        index += 1
        return True, index


class ConstantExpressionParser:
    """Parses a constant expression that can be used in preprocessor statements.

    ```bnf
    <expression> ::= <term>
            | unary_operator <expression>
            | ( "(" <expression> ")" | <expression> ) ( binary_operator <expression> )*
    <term> ::= string
            | constant
            | identifier
            | identifier '(' [ <expression> ("," <expression> )* ] ')'
    ```
    The `string`, `constant` and `identifier` comes from the tokenizer.
    """

    def __init__(self, directive, context, index):
        self.directive = directive
        self.context = context
        self.index = index

    def parse(self):
        try:
            index = self.index
            self.parse_constant_expression()
            if index == self.index:  # No tokens were parsed
                raise CParsingError(f"No argument for #{self.directive} statement")
            if self.context.peek_token(self.index) is None:
                raise CParsingError("Unexpected end of file while parsing constant expression")
            self.index = self.context.skip_ws(self.index, comment=True)
            if not self.context.check_token(self.index, "NEWLINE"):
                raise CParsingError("Unexpected tokens after the constant expression")
            # self.index += 1  # Skip the newline
        except RecursionError:
            raise CParsingError("Constant expression too complex")
        return True, self.index

    def skip_ws(self):
        self.index = self.context.skip_ws(self.index)

    @recursion_limit(100)
    def parse_constant_expression(self):
        self.parse_expression()

    def parse_expression(self):
        if self.context.check_token(self.index, "LPARENTHESIS"):
            self.index += 1
            self.parse_expression()
            if not self.context.check_token(self.index, "RPARENTHESIS"):
                raise CParsingError("Missing closing parenthesis while parsing constant expression")
            self.index += 1
            self.parse_potential_binary_operator()
            return

        if (
                self.context.check_token(self.index, UNARY_OPERATORS)
                or (
                    self.context.check_token(self.index, "IDENTIFIER")
                    and self.context.peek_token(self.index).value == "defined"
                )
        ):
            self.index += 1
            self.skip_ws()
            self.parse_expression()
            return

        if self.context.check_token(self.index, "IDENTIFIER"):
            self.index += 1
            if self.context.check_token(self.index, "LPARENTHESIS"):
                self.index += 1
                self.parse_function_macro()
                return
            self.parse_potential_binary_operator()
            return

        if self.context.check_token(self.index, ("STRING", "CONSTANT", "CHAR_CONST")):
            self.index += 1
            self.parse_potential_binary_operator()
            return

    def parse_function_macro(self):
        self.skip_ws()
        if not self.context.check_token(self.index, "RPARENTHESIS"):
            self.parse_expression()
            while self.context.check_token(self.index, "COMMA"):
                self.index += 1
                self.skip_ws()
                self.parse_expression()
        if not self.context.check_token(self.index, "RPARENTHESIS"):
            raise CParsingError("Missing closing parenthesis")
        self.index += 1
        self.skip_ws()

    def parse_potential_binary_operator(self):
        self.skip_ws()
        if self.context.check_token(self.index, BINARY_OPERATORS):
            self.index += 1
            self.skip_ws()
            self.parse_expression()
            return



errors = {
    "SPC_INSTEAD_TAB": "Spaces at beginning of line",
    "TAB_INSTEAD_SPC": "Found tab when expecting space",
    "CONSECUTIVE_SPC": "Two or more consecutives spaces",
    "CONSECUTIVE_WS": "Two or more consecutives white spaces",
    "SPC_BFR_OPERATOR": "missing space before operator",
    "SPC_AFTER_OPERATOR": "missing space after operator",
    "NO_SPC_BFR_OPR": "extra space before operator",
    "NO_SPC_AFR_OPR": "extra space after operator",
    "SPC_AFTER_PAR": "Missing space after parenthesis (brace/bracket)",
    "SPC_BFR_PAR": "Missing space before parenthesis (brace/bracket)",
    "NO_SPC_AFR_PAR": "Extra space after parenthesis (brace/bracket)",
    "NO_SPC_BFR_PAR": "Extra space before parenthesis (brace/bracket)",
    "SPC_AFTER_POINTER": "space after pointer",
    "SPC_LINE_START": "Unexpected space/tab at line start",
    "SPC_BFR_POINTER": "bad spacing before pointer",
    "SPACE_BEFORE_FUNC": "space before function name",
    "TOO_MANY_TABS_FUNC": "extra tabs before function name",
    "TOO_MANY_TABS_TD": "extra tabs before typedef name",
    "MISSING_TAB_FUNC": "missing tab before function name",
    "MISSING_TAB_VAR": "missing tab before variable name",
    "MISSING_TAB_TYPDEF": "Missing tab before typedef name",
    "TOO_MANY_TAB_VAR": "extra tab before variable name",
    "LINE_TOO_LONG": "line too long",
    "EXP_PARENTHESIS": "Expected parenthesis",
    "MISSING_IDENTIFIER": "missing type qualifier or identifier in function arguments",
    "FORBIDDEN_CHAR_NAME": "user defined identifiers should contain only lowercase characters, \
digits or '_'",
    "TOO_FEW_TAB": "Missing tabs for indent level",
    "TOO_MANY_TAB": "Extra tabs for indent level",
    "TOO_MANY_WS": "Extra whitespaces for indent level",
    "SPACE_REPLACE_TAB": "Found space when expecting tab",
    "TAB_REPLACE_SPACE": "Found tab when expecting space",
    "TOO_MANY_LINES": "Function has more than 25 lines",
    "SPACE_EMPTY_LINE": "Space on empty line",
    "SPC_BEFORE_NL": "Space before newline",
    "TOO_MANY_INSTR": "Too many instructions on a single line",
    "PREPROC_NO_SPACE": "Missing space after preprocessor directive",
    "PREPROC_UKN_STATEMENT": "Unrecognized preprocessor statement",
    "PREPROC_START_LINE": "Preprocessor statement not at the beginning of the line",
    "PREPROC_CONSTANT": "Preprocessor statement must only contain constant defines",
    "PREPROC_EXPECTED_EOL": "Expected EOL after preprocessor statement",
    "PREPROC_BAD_IF": "If preprocessor statement without endif",
    "PREPROC_BAD_ELIF": "Elif preprocessor statement without if or elif",
    "PREPROC_BAD_IFDEF": "Ifdef preprocessor statement without endif",
    "PREPROC_BAD_IFNDEF": "Ifndef preprocessor statement without endif",
    "PREPROC_BAD_ELSE": "Else preprocessor statement without if or elif",
    "PREPROC_BAD_ENDIF": "Endif preprocessor statement without if, elif or else",
    "PREPROC_BAD_INDENT": "Bad preprocessor indentation",
    "PREPROC_MULTLINE": "Multiline preprocessor statement is forbidden",
    "PREPOC_ONLY_GLOBAL": "Preprocessor statements are only allowed in the global scope",
    "USER_DEFINED_TYPEDEF": "User defined typedef must start with t_",
    "STRUCT_TYPE_NAMING": "Structure name must start with s_",
    "ENUM_TYPE_NAMING": "Enum name must start with e_",
    "UNION_TYPE_NAMING": "Union name must start with u_",
    "GLOBAL_VAR_NAMING": "Global variable must start with g_",
    "NO_TAB_BF_TYPEDEF": "Missing whitespace before typedef name",
    "GLOBAL_VAR_DETECTED": "Global variable present in file. Make sure it is a reasonable choice.",
    "EOL_OPERATOR": "Logic operator at the end of line",
    "EMPTY_LINE_FUNCTION": "Empty line in function",
    "EMPTY_LINE_FILE_START": "Empty line at start of file",
    "EMPTY_LINE_FUNCTION": "Empty line in function",
    "EMPTY_LINE_EOF": "Empty line at end of file",
    "WRONG_SCOPE_VAR": "Variable declared in incorrect scope",
    "IMPLICIT_VAR_TYPE": "Missing type in variable declaration",
    "VAR_DECL_START_FUNC": "Variable declaration not at start of function",
    "TOO_MANY_VARS_FUNC": "Too many variables declarations in a function",
    "TOO_MANY_FUNCS": "Too many functions in file",
    "BRACE_SHOULD_EOL": "Expected newline after brace",
    "CONSECUTIVE_NEWLINES": "Consecutive newlines",
    "NEWLINE_PRECEDES_FUNC": "Functions must be separated by a newline",
    "NL_AFTER_VAR_DECL": "Variable declarations must be followed by a newline",
    "NL_AFTER_PREPROC": "Preprocessor statement must be followed by a newline",
    "MULT_ASSIGN_LINE": "Multiple assignations on a single line",
    "MULT_DECL_LINE": "Multiple declarations on a single line",
    "DECL_ASSIGN_LINE": "Declaration and assignation on a single line",
    "FORBIDDEN_CS": "Forbidden control structure",
    "SPACE_AFTER_KW": "Missing space after keyword",
    "RETURN_PARENTHESIS": "Return value must be in parenthesis",
    "EXP_SEMI_COLON": "Expected semicolon",
    "EXP_TAB": "Expected tab",
    "NO_ARGS_VOID": "Empty function argument requires void",
    "MISALIGNED_VAR_DECL": "Misaligned variable declaration",
    "MISALIGNED_FUNC_DECL": "Misaligned function declaration",
    "WRONG_SCOPE_COMMENT": "Comment is invalid in this scope",
    "MACRO_NAME_CAPITAL": "Macro name must be capitalized",
    "MACRO_FUNC_FORBIDDEN": "Macro functions are forbidden",
    "ASSIGN_IN_CONTROL": "Assignment in control structure",
    "VLA_FORBIDDEN": "Variable length array forbidden",
    "TOO_MANY_ARGS": "Function has more than 4 arguments",
    "INCLUDE_HEADER_ONLY": ".c file includes are forbidden",
    "INCLUDE_START_FILE": "Include must be at the start of file",
    "HEADER_PROT_ALL": "Header protection must include all the instructions",
    "HEADER_PROT_ALL_AF": "Instructions after header protection are forbidden",
    "HEADER_PROT_NAME": "Wrong header protection name",
    "HEADER_PROT_UPPER": "Header protection must be in uppercase",
    "HEADER_PROT_MULT": "Multiple header protections, only one is allowed",
    "HEADER_PROT_NODEF": "Header protection not containing #define",
    "TERNARY_FBIDDEN": "Ternaries are forbidden",
    "TOO_MANY_VALS": "Too many values on define",
    "NEWLINE_IN_DECL": "Newline in declaration",
    "MULT_IN_SINGLE_INSTR": "Multiple instructions in single line control structure",
    "NEWLINE_DEFINE": "Newline in define",
    "MISSING_TYPEDEF_ID": "Missing identifier in typedef declaration",
    "LABEL_FBIDDEN": "Label statements are forbidden",
    "GOTO_FBIDDEN": "Goto statements are forbidden",
    "PREPROC_GLOBAL": "Preprocessors can only be used in the global scope",
    "WRONG_SCOPE_FCT": "Function prototype in incorrect scope",
    "WRONG_SCOPE": "Statement is in incorrect scope",
    "INCORRECT_DEFINE": "Incorrect values in define",
    "BRACE_NEWLINE": "Expected newline before brace",
    "EXP_NEWLINE": "Expected newline after control structure",
    "ARG_TYPE_UKN": "Unrecognized variable type",
    "COMMENT_ON_INSTR": "Comment must be on its own line or at end of a line",
    "COMMA_START_LINE": "Comma at line start",
    "MIXED_SPACE_TAB": "Mixed spaces and tabs",
    "ATTR_EOL": "Function attribute must be at the end of line",
    "INVALID_HEADER": "Missing or invalid 42 header",
    "INCLUDE_MISSING_SP": "Missing space between include and filename",
    "TYPE_NOT_GLOBAL": "Enums, structs and unions need to be defined only in global scope",
    "FORBIDDEN_TYPEDEF": "Typedef declaration are not allowed in .c files",
    "FORBIDDEN_STRUCT": "Struct declaration are not allowed in .c files",
    "FORBIDDEN_UNION": "Union declaration are not allowed in .c files",
    "FORBIDDEN_ENUM": "Enum declaration are not allowed in .c files",
    "UNEXPECTED_EOF_CHR": "Unexpected end of file (EOF) while parsing a char",
    "UNEXPECTED_EOL_CHR": "Unexpected end of line (EOL) while parsing a char",
    "UNEXPECTED_EOF_MC": "Unexpected end of file (EOF) while parsing a multiline comment",
    "UNEXPECTED_EOF_STR": "Unexpected end of file (EOF) while parsing a string",
    "EMPTY_CHAR": "Empty character constant",
    "CHAR_AS_STRING": "Character constants can have only one character",
    "INVALID_SUFFIX": "This suffix is invalid",
    "BAD_FLOAT_SUFFIX": "Invalid suffix for float/double literal constant",
    "INVALID_BIN_INT": "Invalid binary integer literal",
    "INVALID_OCT_INT": "Invalid octal integer literal",
    "INVALID_HEX_INT": "Invalid hexadecimal integer literal",
    "MAXIMAL_MUNCH": "Potential maximal munch detected",
    "NO_HEX_DIGITS": "No hexadecimal digits followed by the \\x",
    "UNKNOWN_ESCAPE": "Unknown escape sequence",
    "BAD_EXPONENT": "Exponent has no digits",
    "MULTIPLE_DOTS": "Multiple dots in float constant",
    "MULTIPLE_X": "Multiple 'x' in hexadecimal float constant",
}



from norminette.rules import Rule, Primary


class IsComment(Rule, Primary, priority=90):
    def run(self, context):
        """
        Catches comments tokens
        """
        i = context.skip_ws(0)
        if context.check_token(i, ["MULT_COMMENT", "COMMENT"]) is True:
            self.comment = context.peek_token(i)
            i += 1
            i = context.eol(i)
            return True, i
        return False, 0



from norminette.rules import Rule, Check
from norminette.scope import GlobalScope

ARGUMENTED_PREPROCESSORS = (
    "include",
    "import",
    "if",  # just in case
    None,  # "if" is a special case ...
    "ifdef",
    "ifndef",
    "elif",
    "error",
    "pragma",
    "undef",
    "define",
)


class CheckPreprocessorIndent(Rule, Check):
    depends_on = (
        "IsPreprocessorStatement",
    )

    def run(self, context):
        """
        Preprocessor statements must be indented by an additionnal space for each #ifdef/#ifndef/#if
        statement.
        Structure is `#{indentation}preproc_statement`
        Preprocessor must always be at the start of the line
        Argumented preprocessor statements must have a space between the identifier and the argument
        """
        i = context.skip_ws(0)

        hash_ = context.peek_token(i)
        if hash_ and hash_.line_column != 1:
            context.new_error("PREPROC_START_LINE", hash_)
        if not isinstance(context.scope, GlobalScope):
            context.new_error("PREPOC_ONLY_GLOBAL", hash_)
        i += 1

        # Empty preprocessor statement (only #)
        k = context.skip_ws(i, comment=True)
        if context.check_token(k, "NEWLINE"):
            return

        n = context.skip_ws(i)
        while context.check_token(i, "SPACE"):
            i += 1
        if context.check_token(i, "TAB"):
            context.new_error("TAB_REPLACE_SPACE", context.peek_token(i))
        i = n

        # Check indentation
        spaces = context.peek_token(i).line_column - hash_.line_column - 1
        indent = context.preproc.indent
        if context.check_token(i, ("IF", "ELSE")):
            indent -= 1
        else:
            t = context.peek_token(i)
            if t and t.type == "IDENTIFIER" and t.value.upper() in ("IFNDEF", "IFDEF", "ELIF"):
                indent -= 1
        indent = max(0, indent)
        if spaces > indent:
            context.new_error("TOO_MANY_WS", hash_)
        if spaces < indent:
            context.new_error("PREPROC_BAD_INDENT", hash_)

        # Check spacing after preproc identifier
        if (
            context.check_token(i, ("IDENTIFIER", "IF"))
            and context.peek_token(i).value in ARGUMENTED_PREPROCESSORS
        ):
            i += 1
            # BUG: #error/warning with a "comment" (`#error // Hello`) will be
            # ignored, but it's not a big deal.
            n = context.skip_ws(i, comment=True)
            if context.check_token(n, "NEWLINE"):
                return
            # The idea is to avoid:
            # - `#include"libft.h"`      (no space)
            # - `#include 	 "libft.h"`  (tab)
            # - `#include  "libft.h"`    (two spaces)
            # Note that only `#include "libft.h"` is valid and we also check
            # for `ifdef`, `ifndef`, etc.
            if not context.check_token(i, ("SPACE", "TAB")):
                context.new_error("PREPROC_NO_SPACE", context.peek_token(i))
            j = i
            while context.check_token(i, "SPACE"):
                i += 1
            if context.check_token(i, "TAB"):
                context.new_error("TAB_REPLACE_SPACE", context.peek_token(i))
            if context.skip_ws(j) - j > 1:
                context.new_error("CONSECUTIVE_WS", context.peek_token(j))



from norminette.rules import Rule, Check
import re


class CheckHeader(Rule, Check):
    def parse_header(self, context):
        if context.check_token(0, "MULT_COMMENT") is False:
            context.new_error("INVALID_HEADER", context.peek_token(0))
            context.header_parsed = True
            return
        context.header += context.peek_token(0).value + "\n"

    def check_header(self, context):
        # val = r"\/\* \*{74} \*\/\n\/\*.*\*\/\n\/\*.*\*\/\n\/\*.{3}([^ ]*).*\*\/\n\/\*.*\*\/\n\/\*   By: ([^ ]*).*\*\/\n\/\*.*\*\/\n\/\*   Created: ([^ ]* [^ ]*) by ([^ ]*).*\*\/\n\/\*   Updated: ([^ ]* [^ ]*) by ([^ ]*).*\*\/\n\/\*.*\*\/\n\/\* \*{74} \*\/\n" # noqa: E501
        val_no_check_nl = r"\/\* \*{74} \*\/.\/\*.*\*\/.\/\*.*\*\/.\/\*.{3}([^ ]*).*\*\/.\/\*.*\*\/.\/\*   By: ([^ ]*).*\*\/.\/\*.*\*\/.\/\*   Created: ([^ ]* [^ ]*) by ([^ ]*).*\*\/.\/\*   Updated: ([^ ]* [^ ]*) by ([^ ]*).*\*\/.\/\*.*\*\/.\/\* \*{74} \*\/."  # noqa: E501

        # correct_header = re.match(val, context.header)
        regex = re.compile(val_no_check_nl, re.DOTALL)
        # correct_header_no_nl = re.match(val_no_check_nl, context.header)
        correct_header_no_nl = regex.search(context.header)
        if correct_header_no_nl is None:
            context.new_error("INVALID_HEADER", context.peek_token(0))
        # else:
        #    print (correct_header.group(1,2,3,4,5,6))

    def run(self, context):
        """
        Header checking. Just a warning for now. Does not trigger moulinette error
        """
        if context.header_parsed is True:
            return False, 0
        elif context.history[-1] == "IsComment" and context.header_parsed is False:
            self.parse_header(context)
            context.header_started = True
        elif context.history[-1] != "IsComment" and context.header_started is True:
            self.check_header(context)
            context.header_parsed = True
        elif (
            context.header_started is False
            and context.header_parsed is False
            and context.history[-1] != "IsComment"
        ):
            context.new_error("INVALID_HEADER", context.peek_token(0))
            context.header_parsed = True



from norminette.rules import Rule, Check


class CheckComment(Rule, Check):
    def run(self, context):
        """
        Comments are forbidden inside functions and in the middle of instructions.
        """
        i = context.skip_ws(0)

        tokens = []
        while context.peek_token(i) and not context.check_token(i, "NEWLINE"):
            token = context.peek_token(i)
            tokens.append(token)
            i += 1

        for index, token in enumerate(tokens):
            if token.type in ("COMMENT", "MULT_COMMENT"):
                if self.is_inside_a_function(context):
                    context.new_error("WRONG_SCOPE_COMMENT", token)
                if index == 0 or self.is_last_token(token, tokens[index+1:]):
                    continue
                context.new_error("COMMENT_ON_INSTR", token)

    def is_inside_a_function(self, context):
        if context.history[-2:] == ["IsFuncDeclaration", "IsBlockStart"]:
            return True
        if context.scope.__class__.__name__.lower() == "function":
            return True
        # Sometimes the context scope is a `ControlStructure` scope instead of
        # `Function` scope, so, to outsmart this bug, we need check manually
        # the `context.history`.
        last = None
        for index, record in enumerate(reversed(context.history)):
            if record == "IsFuncDeclaration" and last == "IsBlockStart":
                # Since the limited history API, we can't say if we're in a
                # nested function to reach the first enclosing function, so,
                # we'll consider that the user just declared a normal function
                # in global scope.
                stack = 1
                index -= 1  # Jumps to next record after `IsBlockStart`
                while index > 0 and stack > 0:
                    record = context.history[-index]
                    index -= 1
                    if record not in ("IsBlockStart", "IsBlockEnd"):
                        continue
                    stack = stack + (1, -1)[record == "IsBlockEnd"]
                return bool(stack)
            last = record
        return False

    def is_last_token(self, token, foward):
        expected = ("SPACE", "TAB")
        if token.type == "MULT_COMMENT":
            expected += ("COMMENT", "MULT_COMMENT")
        return all(it.type in ("SPACE", "TAB", "COMMENT", "MULT_COMMENT") for it in foward)



from norminette.rules import Rule, Check


class CheckLabel(Rule, Check):
    def run(self, context):
        """
        Goto and labels are forbidden
        """
        i = 0
        if context.scope.name not in ("Function", "ControlStructure"):
            return False, 0
        i = context.skip_ws(i)
        if context.check_token(i, "GOTO"):
            context.new_error("GOTO_FBIDDEN", context.peek_token(0))
            return False, 0
        if context.check_token(i, "IDENTIFIER") is False:
            return False, 0
        i = context.skip_ws(i + 1)
        if context.check_token(i, "COLON"):
            context.new_error("LABEL_FBIDDEN", context.peek_token(0))
        return False, 0



class NorminetteError(Exception):
    pass


class CParsingError(NorminetteError):
    def __init__(self, errmsg):
        self.msg = errmsg

    def __str__(self):
        return self.msg

    def __repr__(self):
        return self.__str__


class MaybeInfiniteLoop(NorminetteError):
    def __init__(self) -> None:
        super().__init__("The maximum number of iterations a loop can have has been reached")


class UnexpectedEOF(NorminetteError):
    pass



from typing import Tuple, Any

from norminette.context import Context


class Rule:
    __slots__ = ()

    def __new__(cls, context: Context, *args, **kwargs):
        cls.context = context
        cls.name = cls.__name__

        return super().__new__(cls, *args, **kwargs)

    def __repr__(self) -> str:
        return self.name

    def __hash__(self) -> int:
        return hash(self.name)

    def __eq__(self, value: Any) -> bool:
        if isinstance(value, str):
            return self.name == value
        if isinstance(value, Rule):
            return self.name == value.name
        return super().__eq__(value)

    def __ne__(self, value: Any) -> bool:
        return not (self == value)


class Check:
    __slots__ = ()

    depends_on: Tuple[str, ...]

    runs_on_start: bool
    runs_on_rule: bool
    runs_on_end: bool

    def __init_subclass__(cls, **kwargs):
        if not hasattr(cls, "depends_on"):
            cls.depends_on = ()
        cls.runs_on_start = kwargs.pop("runs_on_start", getattr(cls, "runs_on_start", False))
        cls.runs_on_rule = kwargs.pop("runs_on_rule", getattr(cls, "runs_on_rule", not cls.depends_on))
        cls.runs_on_end = kwargs.pop("runs_on_end", getattr(cls, "runs_on_end", False))

    @classmethod
    def register(cls, registry):
        for rule in cls.depends_on:
            registry.dependencies[rule].append(cls)
        if cls.runs_on_start:
            registry.dependencies["_start"].append(cls)
        if cls.runs_on_rule:
            registry.dependencies["_rule"].append(cls)
        if cls.runs_on_end:
            registry.dependencies["_end"].append(cls)

    def is_starting(self):
        """Returns if this `Check` is being run before `Primary`.

        It is only called if `runs_on_start` is set to `True`.
        """
        return self.context.state == "starting"  # type: ignore

    def is_ending(self):
        """Returns if this `Check` is being run after all rules.

        It is only called if `runs_on_end` is set to `True`.
        """
        return self.context.state == "ending"  # type: ignore

    def run(self, context: Context) -> None:
        return


class Primary:
    __slots__ = ()

    priority: int
    scope: Tuple[str, ...]

    def __init_subclass__(cls, **kwargs: Any):
        cls.priority = kwargs.pop("priority", 0)
        if not hasattr(cls, "scope"):
            cls.scope = ()

    def run(self, context: Context) -> Tuple[bool, int]:
        return False, 0



from typing import Optional

red = {
    "TOO_MANY_ARGS",
    "TOO_MANY_VARS_FUNC",
}
yellow = {
    "MIXED_SPACE_TAB",
    "BRACE_NEWLINE"
}
green = {
    "TOO_MANY_FUNCS",
}
blue = {
    "SPC_INSTEAD_TAB",
    "TAB_INSTEAD_SPC",
    "CONSECUTIVE_SPC",
    "CONSECUTIVE_WS",
    "SPC_BFR_OPERATOR",
    "SPC_AFTER_OPERATOR",
    "NO_SPC_BFR_OPR",
    "NO_SPC_AFR_OPR",
    "SPC_AFTER_PAR",
    "SPC_BFR_PAR",
    "NO_SPC_AFR_PAR",
    "NO_SPC_BFR_PAR",
    "SPC_AFTER_POINTER",
    "SPC_LINE_START",
    "SPC_BFR_POINTER",
    "SPACE_BEFORE_FUNC",
    "TOO_MANY_TABS_FUNC",
    "TOO_MANY_TABS_TD",
    "MISSING_TAB_FUNC",
    "MISSING_TAB_VAR",
    "TOO_MANY_TAB_VAR",
    "LINE_TOO_LONG",
    "EXP_PARENTHESIS",
}
pink = {
    "WRONG_SCOPE_COMMENT",
    "COMMENT_ON_INSTR",
}
grey = {
    "INVALID_HEADER",
    "WRONG_SCOPE_COMMENT",
}

_color_table = {
    "91": red,
    "92": green,
    "93": yellow,
    "94": blue,
    "95": pink,
    "97": grey,
}


def error_color(name: str) -> Optional[str]:
    for color, table in _color_table.items():
        if name in table:
            return color
    return None



from norminette.rules import Primary, Rule

cs_keywords = ["DO", "WHILE", "FOR", "IF", "ELSE", "SWITCH", "CASE", "DEFAULT"]
whitespaces = ["TAB", "SPACE", "NEWLINE"]


class IsAmbiguousDeclaration(Primary, Rule, priority=0):
    def run(self, context):
        """
        Catches missing semi-colon or other various missing stuff. Dev feature
        """
        i = context.skip_ws(0, nl=False)
        while context.peek_token(i) and context.check_token(i, "NEWLINE") is False:
            if context.check_token(i, ["SEMI_COLON"]) is False:
                return False, 0
            i += 1
        return True, i



import string

from norminette.rules import Rule, Check
from norminette.scope import GlobalScope, UserDefinedType


assigns = ["ASSIGN"]


class CheckIdentifierName(Rule, Check):
    def run(self, context):
        """
        Function can only be declared in the global scope
        User defined identifiers can only contain lowercase characters, '_' or digits
        """
        legal_characters = string.ascii_lowercase + string.digits + "_"
        if context.history[-1] == "IsFuncDeclaration":
            sc = context.scope
            if type(sc) is not GlobalScope and type(sc) is not UserDefinedType:
                context.new_error("WRONG_SCOPE_FCT", context.peek_token(0))
            while type(sc) != GlobalScope:
                sc = sc.outer()
            for c in sc.fnames[-1]:
                if c not in legal_characters:
                    context.new_error(
                        "FORBIDDEN_CHAR_NAME", context.peek_token(context.fname_pos)
                    )
        if len(context.scope.vars_name) > 0:
            for val in context.scope.vars_name[::]:
                for c in val.value:
                    if c not in legal_characters:
                        context.new_error("FORBIDDEN_CHAR_NAME", val)
                        break
                context.scope.vars_name.remove(val)
        return False, 0



from norminette.rules import Rule, Check


class CheckCommentLineLen(Rule, Check):
    depends_on = ("IsComment",)

    def run(self, context):
        """
        Lines must not be over 80 characters long
        """
        i = 0
        while not context.check_token(i, ["COMMENT", "MULT_COMMENT"]):
            i += 1
        token = context.peek_token(i)
        if not token:
            return
        index = token.pos[1]
        if token.type == "MULT_COMMENT":
            lines = token.value.split("\n")
            # We need to add a padding to the first line because the comment
            # can be at the end of a line.
            lines[0] = " " * index + lines[0]
            for lineno, line in enumerate(lines, start=token.pos[0]):
                if len(line) > 81:
                    token.pos = (lineno, 1)
                    context.new_error("LINE_TOO_LONG", token)
        elif index + len(token.value) > 81:  # token.type == "COMMENT"
            context.new_error("LINE_TOO_LONG", token)



from norminette.rules import Rule, Check


class CheckPreprocessorDefine(Rule, Check):
    depends_on = (
        "IsPreprocessorStatement",
    )

    def run(self, context):
        """
        Defined names must be in capital letters
        Define can only contain constant values, such as integers and strings
        """
        i = context.skip_ws(0)
        i += 1  # skip HASH
        i = context.skip_ws(i)
        if not context.check_token(i, "IDENTIFIER"):
            return
        if not context.peek_token(i).value == "define":
            return
        if context.preproc.skip_define:
            return
        i += 1  # skip DEFINE
        i = context.skip_ws(i)

        if not context.peek_token(i).value.isupper():
            context.new_error("MACRO_NAME_CAPITAL", context.peek_token(i))
        i += 1  # skip macro name

        if context.check_token(i, "LPARENTHESIS"):
            context.new_error("MACRO_FUNC_FORBIDDEN", context.peek_token(i))
            while not context.check_token(i, "RPARENTHESIS"):
                i += 1
            i += 1
        i = context.skip_ws(i)

        # It is obscure what `#define` can hold in its value, see:
        # - https://github.com/42School/norminette/issues/12
        # - https://github.com/42School/norminette/issues/127
        # - https://github.com/42School/norminette/issues/282
        #
        if context.check_token(i, ("MINUS", "PLUS", "BWISE_NOT")):
            i += 1
            i = context.skip_ws(i)
            if not context.check_token(i, ("CONSTANT", "IDENTIFIER")):
                context.new_error("PREPROC_CONSTANT", context.peek_token(i))
                return
            i += 1
        elif context.check_token(i, ("CONSTANT", "IDENTIFIER", "STRING", "CHAR_CONST")):
            i += 1

        i = context.skip_ws(i, comment=True)
        if context.peek_token(i) and not context.check_token(i, "NEWLINE"):
            context.new_error("PREPROC_CONSTANT", context.peek_token(i))



import re
import string
from typing import Optional, Tuple, cast

from norminette.exceptions import UnexpectedEOF, MaybeInfiniteLoop
from norminette.lexer.dictionary import digraphs, trigraphs
from norminette.lexer.dictionary import brackets
from norminette.lexer.dictionary import keywords
from norminette.lexer.dictionary import operators
from norminette.lexer.tokens import Token
from norminette.file import File
from norminette.errors import Error, Highlight as H


def c(a: str, b: str):
    a = a.lower()
    b = b.lower()
    return (
        a + b, a.upper() + b, a + b.upper(), a.upper() + b.upper(),
        b + a, b.upper() + a, b + a.upper(), b.upper() + a.upper(),
    )


quote_prefixes = (*"lLuU", "u8")
octal_digits = "01234567"
hexadecimal_digits = "0123456789abcdefABCDEF"
integer_suffixes = (
    '',
    *"uUlLzZ",
    "ll", "LL",
    "wb", "WB",
    "i64", "I64",
    *c('u', 'l'),
    *c('u', "ll"),
    *c('u', 'z'),
    *c('u', "wb"),
    *c('u', "i64"),
)
float_suffixes = (
    '',
    *"lLfFdD",
    "dd", "DD",
    "df", "DF",
    "dl", "DL",
    *c('f', 'i'),
    *c('f', 'j'),
)

INT_LITERAL_PATTERN = re.compile(r"""
^
# (?P<Sign>[-+]*)
(?P<Prefix>         # prefix can be
    0[bBxX]*        #   0, 0b, 0B, 0x, 0X, 0bb, 0BB, ...
    |               # or empty
)
(?P<Constant>
    # BUG If prefix is followed by two or more x, it doesn't works correctly
    (?<=0[xX])       # is prefix for hex digits?
        [\da-fA-F]+  #   so, collect hex digits
    |                # otherwise
    \d+              #   collect decimal digits
)
(?P<Suffix>
    (?<=[eE])        # is constant ending with an `E`?
        [\w\d+\-.]*  #   so, collect `+` and `-` operators
    |                # otherwise
        \w           #   collect suffixes that starts with an letter
        [\w\d.]*     #     and letters, digits and dots that follows it
    |                # finally, do suffix be optional (empty)
)
""", re.VERBOSE)


def _float_pattern(const: str, digit: str, exponent: Tuple[str, str]):
    pattern = r"""
    ^
    (?P<Constant>{0})
    (?P<Exponent>
        (?:
        [{2}]+[-+]{3}+
        |[{2}]+{3}+
        |(?:[{2}][+-]?(?:[.{3}]+)?)+
        ){1}
    )
    (?P<Suffix>[\w\d._]*|)
    """.format(const, *exponent, digit)
    return re.compile(pattern, re.VERBOSE)


FLOAT_EXPONENT_LITERAL_PATTERN = _float_pattern(r"\d+", digit=r"\d", exponent=('', "eE"))
FLOAT_FRACTIONAL_LITERAL_PATTERN = _float_pattern(r"(?:\d+)?\.\d+|\d+\.", digit=r"\d", exponent=('?', "eE"))
FLOAT_HEXADECIMAL_LITERAL_PATTERN = _float_pattern(r"0[xX]+[\da-fA-F]+(?:\.[\da-fA-F]+)?",
                                                   digit=r"[\da-fA-F]", exponent=('?', "pP"))


class Lexer:
    def __init__(self, file: File):
        self.file = file

        self.__pos = int(0)
        self.__line_pos = self.__line = 1

    def raw_peek(self, *, offset: int = 0, collect: int = 1):
        if (pos := self.__pos + offset) < len(self.file.source):
            return ''.join(self.file.source[pos:pos+collect])
        return None

    def peek(self, *, times: int = 1, offset: int = 0) -> Optional[Tuple[str, int]]:
        char, size = '', 0
        for _ in range(times):
            if (trigraph := self.raw_peek(offset=offset + size, collect=3)) in trigraphs:
                char += trigraphs[trigraph]
                size += 3
            elif (digraph := self.raw_peek(offset=offset + size, collect=2)) in digraphs:
                char += digraphs[digraph]
                size += 2
            elif word := self.raw_peek(offset=offset + size):
                char += word
                size += 1
            else:
                break
        if size:
            return char, size
        return None  # Let it crash :D

    def pop(
        self,
        *,
        times: int = 1,
        use_spaces: bool = False,
        use_escape: bool = False,
    ) -> str:
        result = ""
        for _ in range(times):
            for _ in range(100):
                if peek := self.peek():
                    char, size = peek
                else:
                    raise UnexpectedEOF()
                if char != '\\':
                    break
                peek = self.peek(offset=size)
                if peek is None:
                    break
                temp, _ = peek  # Don't change the `temp` to `char`
                if temp != '\n':
                    if use_escape:
                        if temp in r"abefnrtv\"'?":
                            size += 1
                            char += temp
                        elif temp == 'x':
                            size += 1
                            char += temp
                            # BUG It is just considering one `byte` (0x0 to 0xFF), so it not works correctly
                            # with prefixed strings like `L"\0x1234"`.
                            peek = self.raw_peek(offset=size, collect=2)
                            if peek is None or peek[0] not in hexadecimal_digits:
                                error = Error.from_name("NO_HEX_DIGITS", level="Notice")
                                error.add_highlight(self.__line, self.__line_pos + size - 1, length=1)
                                self.file.errors.add(error)
                            else:
                                for digit in peek:
                                    if digit not in hexadecimal_digits:
                                        break
                                    size += 1
                                    char += digit
                        elif temp in octal_digits:
                            while (temp := self.raw_peek(offset=size)) and temp in octal_digits:
                                size += 1
                                char += temp
                        else:
                            error = Error.from_name("UNKNOWN_ESCAPE", level="Notice")
                            error.add_highlight(self.__line, self.__line_pos + size, length=1)
                            self.file.errors.add(error)
                            char += temp
                            size += 1
                    break
                self.__pos += size + 1
                self.__line += 1
                self.__line_pos = 0
                peek = self.peek()
                if peek is None:
                    raise UnexpectedEOF()
                char, size = peek
            else:
                # It hits when we have multiple lines followed by `\`, e.g:
                # ```c
                # // hello \
                # a \
                #  b \
                # c\
                # \
                # a
                # ```
                raise MaybeInfiniteLoop()
            if char == '\n':
                self.__line_pos = 0
                self.__line += 1
            if char == '\t':
                self.__line_pos += (spaces := 4 - (self.__line_pos - 1) % 4) - 1
                if use_spaces:
                    char = ' ' * spaces
            self.__line_pos += size
            self.__pos += size
            result += char
        return result

    def line_pos(self):
        return self.__line, self.__line_pos

    def parse_char_literal(self) -> Optional[Token]:
        pos = lineno, column = self.line_pos()
        value = ''
        for prefix in quote_prefixes:
            length = len(prefix)
            result = self.raw_peek(collect=length + 1)
            if not result:
                return
            if result.startswith(prefix) and result.endswith('\''):
                value += self.pop(times=length)
                break
        if self.raw_peek() != '\'':
            return
        chars = 0
        value += self.pop()
        for _ in range(100):
            try:
                char = self.pop(use_escape=True)
            except UnexpectedEOF:
                error = Error.from_name("UNEXPECTED_EOF_CHR", highlights=[
                    H(lineno, column, length=len(value)),
                ])
                self.file.errors.add(error)
                break
            if char == '\n':
                error = Error.from_name("UNEXPECTED_EOL_CHR", highlights=[
                    H(lineno, column, length=len(value)),
                    H(lineno, column + len(value), length=1, hint="Perhaps you forgot a single quote (')?")
                ])
                self.file.errors.add(error)
                break
            value += char
            if char == '\'':
                break
            chars += 1
        else:
            raise MaybeInfiniteLoop()
        if value == "''":
            error = Error.from_name("EMPTY_CHAR", highlights=[H(*pos, length=2)])
            self.file.errors.add(error)
        if chars > 1 and value.endswith('\''):
            error = Error.from_name("CHAR_AS_STRING", highlights=[
                H(*pos, length=len(value)),
                H(*pos, length=1,
                  hint="Perhaps you want a string (double quote, \") instead of a char (single quote, ')?"),
            ])
            self.file.errors.add(error)
        return Token("CHAR_CONST", pos, value=value)

    def parse_string_literal(self):
        """String constants can contain any characer except unescaped newlines.
        An unclosed string or unescaped newline is a fatal error and thus
        parsing will stop here.
        """
        if not self.peek():
            return
        pos = lineno, column = self.line_pos()
        val = ''
        for prefix in quote_prefixes:
            length = len(prefix)
            result = self.raw_peek(collect=length + 1)
            if not result:
                return
            if result.startswith(prefix) and result.endswith('"'):
                val += self.pop(times=length)
                break
        if self.raw_peek() != '"':
            return
        val += self.pop()
        while self.peek() is not None:
            char = self.pop(use_escape=True)
            val += char
            if char == '"':
                break
        else:
            error = Error.from_name("UNEXPECTED_EOF_STR")
            error.add_highlight(*pos, length=len(val))
            error.add_highlight(lineno, column + len(val), length=1, hint="Perhaps you forgot a double quote (\")?")
            self.file.errors.add(error)
        return Token("STRING", pos, val)

    def parse_integer_literal(self):
        # TODO Add to support single quote (') to separate digits according to C23

        match = INT_LITERAL_PATTERN.match(self.file.source[self.__pos:])
        if match is None:
            return

        pos = lineno, column = self.line_pos()
        token = Token("CONSTANT", pos, slice := self.pop(times=match.end()))

        if match["Suffix"] not in integer_suffixes:
            suffix_length = len(match["Suffix"])
            string_length = len(slice) - suffix_length
            if match["Suffix"][0] in "+-":
                error = Error.from_name("MAXIMAL_MUNCH")
                error.add_highlight(lineno, column + string_length, length=1, hint="Perhaps you forgot a space ( )?")
            else:
                error = Error.from_name("INVALID_SUFFIX")
                error.add_highlight(lineno, column + string_length, length=suffix_length)
            self.file.errors.add(error)

        def _check_bad_prefix(name: str, bucket: str):
            error = Error.from_name(f"INVALID_{name}_INT")
            for index, char in enumerate(match["Constant"], start=len(match["Prefix"])):
                if char not in bucket:
                    error.add_highlight(lineno, column + index, length=1)
            if error.highlights:
                self.file.errors.add(error)

        if match["Prefix"] in ("0b", "0B"):
            _check_bad_prefix("BIN", "01")
        elif match["Prefix"] == '0':
            _check_bad_prefix("OCT", "01234567")
        elif match["Prefix"] in ("0x", "0X"):
            _check_bad_prefix("HEX", "0123456789abcdefABCDEF")

        return token

    def parse_float_literal(self):
        """Numeric constants can take many forms:
        - integer constants only allow digits [0-9]
        - real number constant only allow digits [0-9],
            ONE optionnal dot '.' and ONE optionnal 'e/E' character
        - binary constant only allow digits [0-1] prefixed by '0b' or '0B'
        - hex constant only allow digits [0-9], letters [a-f/A-F] prefixed
            by '0x' or '0X'
        - octal constants allow digits [0-9] prefixed by a zero '0'
            character

        Size ('l/L' for long) and sign ('u/U' for unsigned) specifiers can
        be appended to any of those. tokens

        Plus/minus operators ('+'/'-') can prefix any of those tokens

        a numeric constant could start with a '.' (dot character)
        """
        constant = self.raw_peek()
        if constant is None:
            return
        pos = lineno, column = self.line_pos()
        src = self.file.source[self.__pos:]
        if match := FLOAT_EXPONENT_LITERAL_PATTERN.match(src):
            type = "exponent"
        elif match := FLOAT_FRACTIONAL_LITERAL_PATTERN.match(src):
            type = "fractional"
        elif match := FLOAT_HEXADECIMAL_LITERAL_PATTERN.match(src):
            type = "hexadecimal"
        else:
            return
        error = None
        suffix = len(match["Suffix"])
        column += len(match["Constant"])
        badhex = match["Constant"].strip(hexadecimal_digits + '.')
        if type == "exponent" and not re.match(r"[eE][-+]?\d+", match["Exponent"]):
            error = Error.from_name("BAD_EXPONENT")
            error.add_highlight(lineno, column, length=len(match["Exponent"]) + suffix)
        elif type == "hexadecimal" and '.' not in match["Constant"] and not match["Exponent"]:
            return  # Hexadecimal Integer
        elif type == "hexadecimal" and badhex not in ('x', 'X'):
            error = Error.from_name("MULTIPLE_X")
            error.add_highlight(lineno, column - len(match["Constant"]) + 1, length=len(badhex))
        elif match["Constant"].count('.') == 1 and match["Suffix"].count('.') > 0:
            error = Error.from_name("MULTIPLE_DOTS")
            error.add_highlight(lineno, column, length=len(match["Exponent"]) + suffix)
        elif match["Suffix"] not in float_suffixes:
            error = Error.from_name("BAD_FLOAT_SUFFIX")
            error.add_highlight(lineno, column + len(match["Exponent"]), length=suffix)
        if error:
            self.file.errors.add(error)
        return Token("CONSTANT", pos, self.pop(times=match.end()))

    def parse_multi_line_comment(self) -> Optional[Token]:
        if self.raw_peek(collect=2) != "/*":
            return
        pos = lineno, column = self.line_pos()
        val = self.pop(times=2)
        eof = False
        while self.peek():
            try:
                val += self.pop(use_spaces=True)
            except UnexpectedEOF:
                eof = True
                break
            if val.endswith("*/"):
                break
        else:
            eof = True
        if eof:
            # TODO Add a better highlight since it is a multi-line token
            error = Error.from_name("UNEXPECTED_EOF_MC")
            error.add_highlight(lineno, column, length=len(val))
            self.file.errors.add(error)
        return Token("MULT_COMMENT", pos, val)

    def parse_line_comment(self) -> Optional[Token]:
        """Comments are anything after '//' characters, up until a newline or
        end of file
        """
        if self.raw_peek(collect=2) != "//":
            return
        pos = self.line_pos()
        val = self.pop(times=2)
        while result := self.peek():
            char, _ = result
            if char == '\n':
                break
            try:
                val += self.pop()
            except UnexpectedEOF:
                break
        return Token("COMMENT", pos, val)

    def parse_identifier(self) -> Optional[Token]:
        """Identifiers can start with any letter [a-z][A-Z] or an underscore
        and contain any letters [a-z][A-Z] digits [0-9] or underscores
        """
        char = self.raw_peek()
        if not char or char not in string.ascii_letters + '_':
            return
        pos = self.line_pos()
        val = self.pop()
        while char := self.raw_peek():
            if char not in string.ascii_letters + "0123456789_":
                break
            val += self.pop()
        if val in keywords:
            return Token(keywords[val], pos)
        return Token("IDENTIFIER", pos, val)

    def parse_operator(self):
        """Operators can be made of one or more sign, so the longest operators
        need to be looked up for first in order to avoid false positives
        eg: '>>' being understood as two 'MORE_THAN' operators instead of
            one 'RIGHT_SHIFT' operator
        """
        result = self.peek()
        if not result:
            return
        char, _ = result
        if char not in "+-*/,<>^&|!=%;:.~?#":
            return
        pos = self.line_pos()
        if char in ".+-*/%<>^&|!=":
            if self.raw_peek(collect=3) in (">>=", "<<=", "..."):
                return Token(operators[self.pop(times=3)], pos)
            temp, _ = self.peek(times=2)  # type: ignore
            if temp in (">>", "<<", "->"):
                return Token(operators[self.pop(times=2)], pos)
            if temp == char + "=":
                return Token(operators[self.pop(times=2)], pos)
            if char in "+-<>=&|":
                if temp == char * 2:
                    return Token(operators[self.pop(times=2)], pos)
        char = self.pop()
        return Token(operators[char], pos)

    def parse_whitespace(self) -> Optional[Token]:
        char = self.raw_peek()
        if char is None or char not in "\n\t ":
            return
        if char == ' ':
            token = Token("SPACE", self.line_pos())
        elif char == "\t":
            token = Token("TAB", self.line_pos())
        elif char == "\n":
            token = Token("NEWLINE", self.line_pos())
        self.pop()
        return token

    def parse_brackets(self) -> Optional[Token]:
        result = self.peek()
        if result is None:
            return
        char, _ = result
        if char not in brackets:
            return
        start = self.line_pos()
        value = self.pop()
        return Token(brackets[value], start)

    parsers = (
        parse_float_literal,  # Need to be above:
                              #  `parse_operator` to avoid `<DOT>`
                              #  `parse_integer_literal` to avoid `\d+`
        parse_integer_literal,
        parse_char_literal,
        parse_string_literal,
        parse_identifier,  # Need to be bellow `char` and `string`
        parse_whitespace,
        parse_line_comment,
        parse_multi_line_comment,
        parse_operator,
        parse_brackets,
    )

    def get_next_token(self) -> Optional[Token]:
        """Peeks one character and tries to match it to a token type,
        if it doesn't match any of the token types, an error will be raised
        and current file's parsing will stop
        """
        while self.raw_peek():
            if self.raw_peek(collect=2) == "\\\n" or self.raw_peek(collect=4) == "??/\n":
                # Avoid using `.pop()` here since it ignores the escaped
                # newline and pops and upcomes after it. E.g, if we have
                # `\\\nab` and use `.pop()`, the parsers funcs will see `b``.
                _, size = self.peek()  # type: ignore
                self.__pos += cast(int, size) + 1
                self.__line += 1
                self.__line_pos = 1
            else:
                break
        for parser in self.parsers:
            if result := parser(self):
                return result
        if char := self.raw_peek():
            error = Error("BAD_LEXEME", f"No matchable token for '{char}' lexeme")
            error.add_highlight(*self.line_pos(), length=1)
            self.file.errors.add(error)
            self.__pos += 1
            self.__line_pos += 1
            # BUG If we have multiples bad lexemes, it can raise RecursionError
            return self.get_next_token()

    def __iter__(self):
        while token := self.get_next_token():
            yield token



import os.path
import itertools

from norminette.rules import Rule, Check


class CheckPreprocessorInclude(Rule, Check):
    depends_on = (
        "IsPreprocessorStatement",
    )

    def run(self, context):
        """
        Includes must be at the start of the file
        You cannot include anything that isn't an header file
        """
        i = hash_index = context.skip_ws(0)
        i += 1  # skip HASH
        i = context.skip_ws(i)
        if context.check_token(i, "IDENTIFIER") is False:
            return False, 0
        if context.peek_token(i).value != "include":
            return False, 0
        if not self.is_in_start_of_file(context):
            context.new_error("INCLUDE_START_FILE", context.peek_token(hash_index))

        i += 1  # skip INCLUDE
        i = context.skip_ws(i)
        if context.check_token(i, "STRING"):  # "niumxp.h"
            file = context.peek_token(i).value.strip().strip('"')
            file, extension = os.path.splitext(file)
            if extension != ".h":
                context.new_error("INCLUDE_HEADER_ONLY", context.peek_token(i))
        else:  # <niumxp.h>
            less = context.peek_token(i)
            while not context.check_token(i, "MORE_THAN"):
                i += 1
            last = context.peek_token(i - 1)
            prev = context.peek_token(i - 2)
            if last.type != "IDENTIFIER" or not (last.value == "h" and prev.type == "DOT"):
                context.new_error("INCLUDE_HEADER_ONLY", less)

        i += 1  # skip MORE_THAN or STRING

        return True, i

    def is_in_start_of_file(self, context):
        """Check if the include is at the start of the file
        """
        headers = (
            "IsComment",
            "IsEmptyLine",
            "IsPreprocessorStatement",
        )
        history = itertools.filterfalse(lambda item: item in headers, context.history)
        return (
            context.scope.include_allowed
            and next(history, None) is None
        )



from norminette.rules import Rule, Check
from norminette.scope import GlobalScope, ControlStructure


class CheckBlockStart(Rule, Check):
    depends_on = (
        "IsBlockStart",
    )

    def run(self, context):
        """
        Braces signal that the control structure, function, or user defined type can contain
        multiple lines.
        A control structure that has no braces can only contain one instruction line, but can
        contain multiple control structures
        """
        outer = context.scope.get_outer()
        if len(context.history) > 2:
            i = 0
            for item in context.history[::]:
                if item != "IsEmptyLine":
                    if i == 2:
                        hist_1 = item
                    elif i == 3:
                        hist_2 = item
                    i += 1
            if (
                type(context.scope) is GlobalScope
                and context.scope.tmp_scope is not None
                and hist_1 == "IsFuncDeclaration"
                and hist_2 == "IsPreprocessorStatement"
            ):
                context.scope.functions -= 1
                context.scope = context.tmp_scope
                context.scope.multiline = True
                context.tmp_scope = None
        if (
            type(context.scope) == ControlStructure
            and outer is not None
            and type(outer) == ControlStructure
        ):
            if outer.multiline is False:
                context.new_error("MULT_IN_SINGLE_INSTR", context.peek_token(0))
        return False, 0



import os
from typing import Optional

from norminette.errors import Errors


class File:
    def __init__(self, path: str, source: Optional[str] = None) -> None:
        self.path = path
        self._source = source

        self.errors = Errors()
        self.basename = os.path.basename(path)
        self.name, self.type = os.path.splitext(self.basename)

    @property
    def source(self) -> str:
        if self._source is None:
            with open(self.path) as file:
                self._source = file.read()
        return self._source

    def __repr__(self) -> str:
        return f"<File {self.path!r}>"



from norminette.rules import Rule, Primary


class IsLabel(Rule, Primary, priority=10):
    def run(self, context):
        """
        Catches label and raises norm error whenever
        """
        i = context.skip_ws(0)
        if context.check_token(i, "IDENTIFIER") is False:
            return False, 0
        i = context.skip_ws(i + 1)  # +1 to skip the identifier
        if context.check_token(i, "COLON") is False:
            return False, 0
        i = context.skip_ws(i + 1)  # +1 to skip the colon
        if context.check_token(i, "NEWLINE"):
            return True, i+1
        while context.peek_token(i) and context.check_token(i, "SEMI_COLON") is False:
            i += 1
        i = context.eol(i + 1)  # +1 to skip the semi-colon
        return True, i



import itertools

from norminette.rules import Rule, Check


class CheckPreprocessorProtection(Rule, Check):
    depends_on = (
        "IsPreprocessorStatement",
    )

    def run(self, context):
        """
        Header protection must be as follows:
        ```c
        #ifndef __FILENAME_H__
        # define __FILENAME_H__
        #endif
        ```
        Any header instruction must be within the header protection
        """
        if context.file.type != ".h":
            return False, 0
        i = context.skip_ws(0)
        hash = context.peek_token(i)
        i += 1  # Skip the HASH
        i = context.skip_ws(i)
        if not context.check_token(i, "IDENTIFIER"):
            return False, 0
        # TODO: Add to check if macro definition is bellow #ifndef
        t = context.peek_token(i)
        if not t or t.type != "IDENTIFIER" or t.value.upper() not in ("IFNDEF", "ENDIF"):
            return False, 0
        i += 1
        guard = context.file.basename.upper().replace(".", "_")
        if t.value.upper() == "ENDIF":
            if context.preproc.indent == 0 and not context.protected:
                i = context.skip_ws(i, nl=True, comment=True)
                if context.peek_token(i) is not None:
                    context.new_error("HEADER_PROT_ALL_AF", context.peek_token(i))
                if not context.preproc.has_macro_defined(guard):
                    context.new_error("HEADER_PROT_NODEF", hash)
                context.protected = True
            return False, 0
        if context.preproc.indent != 1:
            return False, 0
        i = context.skip_ws(i)
        macro = context.peek_token(i).value
        if macro != guard and not context.protected:
            if macro.upper() == guard:
                context.new_error("HEADER_PROT_UPPER", context.peek_token(i))
            else:
                context.new_error("HEADER_PROT_NAME", context.peek_token(i))

        if context.protected:
            context.new_error("HEADER_PROT_MULT", hash)
            return False, 0

        headers = (
            "IsComment",
            "IsEmptyLine",
        )
        history = context.history[:-1]  # Remove the current `IsPreprocessorStatement`
        history = itertools.filterfalse(lambda item: item in headers, history)
        if next(history, None):
            # We can't say what line contains the instruction outside
            # header protection due to limited history information.
            context.new_error("HEADER_PROT_ALL", hash)

        return False, 0



from typing import Optional, Tuple
from dataclasses import dataclass, field


@dataclass(eq=True, repr=True)
class Token:
    type: str
    pos: Tuple[int, int]
    value: Optional[str] = field(default=None)

    @property
    def length(self) -> int:
        return len(self.value or '')

    @property
    def unsafe_length(self) -> Optional[int]:
        if self.value is None:
            return None
        return self.length

    @property
    def lineno(self) -> int:
        return self.pos[0]

    @property
    def column(self) -> int:
        return self.pos[1]

    @property
    def line_column(self):
        return self.pos[1]

    def __str__(self):
        """
        Token representation for debugging, using the format <TYPE=value>
        or simply <TYPE> when value is None
        """
        r = f"<{self.type}={self.value}>" if self.value else f"<{self.type}>"
        return r



from norminette.rules import Rule, Check

types = [
    "INT",
    "VOID",
    "CHAR",
    "LONG",
    "DOUBLE",
    "SHORT",
    "SIGNED",
    "UNSIGNED",
    "STATIC",
    "CONST",
    "IDENTIFIER",
    "ENUM",
    "STRUCT",
    "UNION",
    "VOLATILE",
    "EXTERN",
    "SPACE",
    "TAB",
]


class CheckGlobalNaming(Rule, Check):
    depends_on = (
        "IsVarDeclaration",
    )

    def run(self, context):
        """
        Global variable names must be preceded by g_
        """
        i = 0
        if context.scope.name != "GlobalScope":
            return False, 0
        i = context.skip_ws(i)
        _, i = context.check_type_specifier(i)
        i = context.skip_ws(i)
        while context.check_token(i, "IDENTIFIER") is False:
            i += 1
        if (
            context.peek_token(i) is not None
            and context.peek_token(i).value != "environ"
        ):
            context.new_warning("GLOBAL_VAR_DETECTED", context.peek_token(0))
            if context.peek_token(i).value.startswith("g_") is False:
                context.new_error("GLOBAL_VAR_NAMING", context.peek_token(i))
        return False, i



from norminette.rules import Rule, Check


allowed_in_header = [
    "IsVarDeclaration",
    "IsUserDefinedType",
    "IsPreprocessorStatement",
    "IsEmptyLine",
    "IsBlockStart",
    "IsBlockEnd",
    "IsComment",
    "IsEndOfLine",
    "IsFuncPrototype",
]

must_be_within_define = [
    "IsVarDeclaration",
    "IsUserDefinedType",
    "IsFuncPrototype",
]


class CheckInHeader(Rule, Check):
    depends_on = (
        "IsVarDeclaration",
        "IsUserDefinedType",
        "IsPreprocessorStatement",
        "IsEmptyLine",
        "IsBlockStart",
        "IsBlockEnd",
        "IsComment",
        "IsEndOfLine",
        "IsFuncPrototype",
    )

    def run(self, context):
        """
        Each .h file must be protected against double inclusion
        Instructions allowed in header files:
            - Variable declaration
            - User defined types
            - Comments
            - Function prototypes
        """
        if context.file.type != ".h":
            return False, 0
        sc = context.scope
        while sc.name != "GlobalScope":
            sc = sc.get_outer()
        if context.history[-1] not in allowed_in_header:
            context.new_error("FORBIDDEN_IN_HEADER", context.peek_token(0))
            return False, 0
        return False, 0



from norminette.rules import Rule, Check


class CheckGeneralSpacing(Rule, Check):
    depends_on = (
        "IsDeclaration",
        "IsControlStatement",
        "IsExpressionStatement",
        "IsAssignation",
        "IsFunctionCall",
    )

    def run(self, context):
        """
        Checks for tab/space consistency
        """
        if context.scope.name == "UserDefinedType":
            return False, 0
        i = context.skip_ws(0)
        while i < context.tkn_scope:
            if context.check_token(i, "TAB") is True:
                context.new_error("TAB_INSTEAD_SPC", context.peek_token(i))
                break
            if context.check_token(i, ["NEWLINE", "ESCAPED_NEWLINE"]) is True:
                i = context.skip_ws(i + 1, nl=True)
            i += 1
        return False, 0
